/* 
    JSPWiki - a JSP-based WikiWiki clone.

    Copyright (C) 2001-2005 Janne Jalkanen (Janne.Jalkanen@iki.fi)

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU Lesser General Public License as published by
    the Free Software Foundation; either version 2.1 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU Lesser General Public License for more details.

    You should have received a copy of the GNU Lesser General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */
package org.cs3.dokuwiki.jsp.client.jspparser;

import java.util.ArrayList;
import java.util.Arrays;
import java.util.Collection;
import java.util.Comparator;
import java.util.EmptyStackException;
import java.util.HashMap;
import java.util.Iterator;
import java.util.List;
import java.util.Set;
import java.util.Stack;

import org.cs3.dokuwiki.jsp.client.apache.oro.text.GlobCompiler;
import org.cs3.dokuwiki.jsp.client.apache.oro.text.regex.MalformedPatternException;
import org.cs3.dokuwiki.jsp.client.apache.oro.text.regex.MatchResult;
import org.cs3.dokuwiki.jsp.client.apache.oro.text.regex.Pattern;
import org.cs3.dokuwiki.jsp.client.apache.oro.text.regex.PatternCompiler;
import org.cs3.dokuwiki.jsp.client.apache.oro.text.regex.PatternMatcher;
import org.cs3.dokuwiki.jsp.client.apache.oro.text.regex.Perl5Compiler;
import org.cs3.dokuwiki.jsp.client.apache.oro.text.regex.Perl5Matcher;

import com.google.gwt.xml.client.Document;
import com.google.gwt.xml.client.Element;
import com.google.gwt.xml.client.Node;
import com.google.gwt.xml.client.NodeList;
import com.google.gwt.xml.client.Text;
import com.google.gwt.xml.client.XMLParser;

/**
 *  Parses JSPWiki-style markup into a WikiDocument DOM tree.  This class is the
 *  heart and soul of JSPWiki : make sure you test properly anything that is added,
 *  or else it breaks down horribly.
 *  
 *  Modified by a lot to suite GWT... (pb)
 *
 *  @author Janne Jalkanen, Pascal Bihler
 *  @since  2.4
 */
public class JSPWikiMarkupParser
    extends MarkupParser
{

    public static final String PI_DISABLE_OUTPUT_ESCAPING =
        "javax.xml.transform.disable-output-escaping";
    
    /** The value for anchor element <tt>class</tt> attributes when used 
      * for wiki page (normal) links. The value is "wikipage". */
    public static final String CLASS_WIKIPAGE = "wikipage";

    /** The value for anchor element <tt>class</tt> attributes when used 
      * for edit page links. The value is "editpage". */
    public static final String CLASS_EDITPAGE = "editpage";
    
    /** The value for anchor element <tt>class</tt> attributes when used 
      * for interwiki page links. The value is "interwiki". */
    public static final String CLASS_INTERWIKI = "interwiki";
    
    protected static final int              READ          = 0;
    protected static final int              EDIT          = 1;
    protected static final int              EMPTY         = 2;  // Empty message
    protected static final int              LOCAL         = 3;
    protected static final int              LOCALREF      = 4;
    protected static final int              IMAGE         = 5;
    protected static final int              EXTERNAL      = 6;
    protected static final int              INTERWIKI     = 7;
    protected static final int              IMAGELINK     = 8;
    protected static final int              IMAGEWIKILINK = 9;
    protected static final int              ATTACHMENT    = 10;

   // private static Logger log = Logger.getLogger( JSPWikiMarkupParser.class );

    //private boolean        m_iscode       = false;
    private boolean        m_isbold       = false;
    private boolean        m_isitalic     = false;
    private boolean        m_istable      = false;
    private boolean        m_isPre        = false;
    private boolean        m_isEscaping   = false;
    private boolean        m_isdefinition = false;
    private boolean        m_isPreBlock   = false;

    /** Contains style information, in multiple forms. */
    private Stack          m_styleStack   = new Stack();
    
     // general list handling
    private int            m_genlistlevel = 0;
    private StringBuffer   m_genlistBulletBuffer = new StringBuffer(10);  // stores the # and * pattern
    private boolean        m_allowPHPWikiStyleLists = true;


    private boolean        m_isOpenParagraph = false;

    /** Keeps image regexp Patterns */
    private List           m_inlineImagePatterns;

    private PatternMatcher m_inlineMatcher = new Perl5Matcher();

    /** Keeps track of any plain text that gets put in the Text nodes */
    private StringBuffer   m_plainTextBuf = new StringBuffer(20);
    
    private Element        m_currentElement;
    
    /**
     *  This property defines the inline image pattern.  It's current value
     *  is jspwiki.translatorReader.inlinePattern
     */
    public static final String     PROP_INLINEIMAGEPTRN  = "jspwiki.translatorReader.inlinePattern";

    /** If true, consider CamelCase hyperlinks as well. */
    public static final String     PROP_CAMELCASELINKS   = "jspwiki.translatorReader.camelCaseLinks";

    /** If true, all hyperlinks are translated as well, regardless whether they
        are surrounded by brackets. */
    public static final String     PROP_PLAINURIS        = "jspwiki.translatorReader.plainUris";

    /** If true, all outward links (external links) have a small link image appended. */
    public static final String     PROP_USEOUTLINKIMAGE  = "jspwiki.translatorReader.useOutlinkImage";

    /** If true, all outward attachment info links have a small link image appended. */
    public static final String     PROP_USEATTACHMENTIMAGE = "jspwiki.translatorReader.useAttachmentImage";
    
    /** If set to "true", all external links are tagged with 'rel="nofollow"' */
    public static final String     PROP_USERELNOFOLLOW   = "jspwiki.translatorReader.useRelNofollow";

    /** If true, then considers CamelCase links as well. */
    private boolean                m_camelCaseLinks      = false;

    /** If true, consider URIs that have no brackets as well. */
    // FIXME: Currently reserved, but not used.
    private boolean                m_plainUris           = false;

    /** If true, all outward links use a small link image. */
    //private boolean                m_useOutlinkImage     = true;

    //private boolean                m_useAttachmentImage  = true;

    /** If true, allows raw HTML. */
    private boolean                m_allowHTML           = false;

    private boolean                m_useRelNofollow      = false;

    private PatternCompiler        m_compiler = new Perl5Compiler();

    static final String WIKIWORD_REGEX = "(^|[[:^alnum:]]+)([[:upper:]]+[[:lower:]]+[[:upper:]]+[[:alnum:]]*|(http://|https://|mailto:)([A-Za-z0-9_/\\.\\+\\?\\#\\-\\@=&;]+))";
    
    private PatternMatcher         m_camelCaseMatcher = new Perl5Matcher();
    private Pattern                m_camelCasePattern;
    
    private int                    m_rowNum              = 1;
    
    private HashMap patternDatabase = new HashMap();
    

    /**
     *  The default inlining pattern.  Currently "*.png"
     */
    public static final String     DEFAULT_INLINEPATTERN = "*.png";

    /**
     *  This list contains all IANA registered URI protocol
     *  types as of September 2004 + a few well-known extra types.
     *
     *  JSPWiki recognises all of them as external links.
     *  
     *  This array is sorted during class load, so you can just dump
     *  here whatever you want in whatever order you want.
     */
    static final String[] c_externalLinks = {
        "http:", "ftp:", "https:", "mailto:",
        "news:", "file:", "rtsp:", "mms:", "ldap:",
        "gopher:", "nntp:", "telnet:", "wais:",
        "prospero:", "z39.50s", "z39.50r", "vemmi:",
        "imap:", "nfs:", "acap:", "tip:", "pop:",
        "dav:", "opaquelocktoken:", "sip:", "sips:",
        "tel:", "fax:", "modem:", "soap.beep:", "soap.beeps",
        "xmlrpc.beep", "xmlrpc.beeps", "urn:", "go:",
        "h323:", "ipp:", "tftp:", "mupdate:", "pres:",
        "im:", "mtqp", "smb:" };

    private static final String INLINE_IMAGE_PATTERNS = "JSPWikiMarkupParser.inlineImagePatterns";

    private static final String CAMELCASE_PATTERN     = "JSPWikiMarkupParser.camelCasePattern";

    private static String[] CLASS_TYPES =
    {
       CLASS_WIKIPAGE,
       CLASS_EDITPAGE,
       "",
       "footnote",
       "footnoteref",
       "",
       "external",
       CLASS_INTERWIKI,
       "external",
       CLASS_WIKIPAGE,
       "attachment"
    };
    

    /**
     *  This Comparator is used to find an external link from c_externalLinks.  It
     *  checks if the link starts with the other arraythingie.
     */
    private static Comparator c_startingComparator = new StartingComparator();

    static
    {
        Arrays.sort( c_externalLinks );
    }
    
    /**
     *  Creates a markup parser.
     * @param camelCaseLinks 
     * @param plainUris 
     * @param useOutlinkImage 
     * @param useAttachmentImage 
     * @param allowHTML 
     * @param useRelNofollow 
     */
    public JSPWikiMarkupParser(  Reader in, boolean camelCaseLinks, boolean plainUris, boolean allowHTML, boolean useRelNofollow )
    {
        super( in );
        //Set Properties
        m_camelCaseLinks  = camelCaseLinks;

        m_plainUris           = plainUris;
        m_allowHTML           = allowHTML;
        m_useRelNofollow      = useRelNofollow;
        
        initialize();
        
    
    }
    public JSPWikiMarkupParser(  Reader in )
    {
    	this(in,false,false,true,false);
    }

    /**
     *  @param engine The WikiEngine this reader is attached to.  Is
     * used to figure out of a page exits.
     */

    // FIXME: parsers should be pooled for better performance.
    private void initialize()
    {
        PatternCompiler compiler         = new GlobCompiler();
        List            compiledpatterns;

        //
        //  We cache compiled patterns in the engine, since their creation is
        //  really expensive
        //
        compiledpatterns = (List)patternDatabase.get( INLINE_IMAGE_PATTERNS );
        
        if( compiledpatterns == null )
        {
            compiledpatterns = new ArrayList();
            Collection ptrns = getImagePatterns( patternDatabase );

            //
            //  Make them into Regexp Patterns.  Unknown patterns
            //  are ignored.
            //
            for( Iterator i = ptrns.iterator(); i.hasNext(); )
            {
                try
                {       
                    compiledpatterns.add( compiler.compile( (String)i.next(), 
                                                            GlobCompiler.DEFAULT_MASK|GlobCompiler.READ_ONLY_MASK ) );
                }
                catch( MalformedPatternException e )
                {
                    //log.error("Malformed pattern in properties: ", e );
                }
            }
            
            patternDatabase.put( INLINE_IMAGE_PATTERNS, compiledpatterns );
        }
        
        //m_inlineImagePatterns = Collections.unmodifiableList(compiledpatterns);
        m_inlineImagePatterns = compiledpatterns;

        m_camelCasePattern = (Pattern) patternDatabase.get( CAMELCASE_PATTERN );
        if( m_camelCasePattern == null )
        {
            try
            {
                m_camelCasePattern = m_compiler.compile( WIKIWORD_REGEX, 
                                                         Perl5Compiler.DEFAULT_MASK|Perl5Compiler.READ_ONLY_MASK );
            }
            catch( MalformedPatternException e )
            {
                //log.fatal("Internal error: Someone put in a faulty pattern.",e);
                //throw new InternalWikiException("Faulty camelcasepattern in TranslatorReader");
            }
            patternDatabase.put( CAMELCASE_PATTERN, m_camelCasePattern );
        }

      
        disableAccessRules();
        
    }


    /**
     *  Figure out which image suffixes should be inlined.
     *  @return Collection of Strings with patterns.
     */

    // FIXME: Does not belong here; should be elsewhere
    public static Collection getImagePatterns( HashMap engine )
    {
        Set keys = engine.keySet();
        ArrayList  ptrnlist = new ArrayList();
        
        for(Iterator it = keys.iterator(); it.hasNext(); )
        {
            String name = (String) it.next();

            if( name.startsWith( PROP_INLINEIMAGEPTRN ) )
            {
                String ptrn = (String) engine.get(name);

                ptrnlist.add( ptrn );
            }
        }

        if( ptrnlist.size() == 0 )
        {
            ptrnlist.add( DEFAULT_INLINEPATTERN );
        }

        return ptrnlist;
    }

    /**
     *  Calls a transmutator chain.
     *
     *  @param list Chain to call
     *  @param text Text that should be passed to the mutate() method
     *              of each of the mutators in the chain.
     *  @return The result of the mutation.
     */

    protected String callMutatorChain( Collection list, String text )
    {
        if( list == null || list.size() == 0 )
        {
            return text;
        }

        for( Iterator i = list.iterator(); i.hasNext(); )
        {
            StringTransmutator m = (StringTransmutator) i.next();

            text = m.mutate( text );
        }

        return text;
    }

    /**
     * Calls the heading listeners.
     * 
     * @param param A Heading object.
     */
    protected void callHeadingListenerChain( Heading param )
    {
        List list = m_headingListenerChain;

        for( Iterator i = list.iterator(); i.hasNext(); )
        {
            HeadingListener h = (HeadingListener) i.next();
        }
    }

    /**
     *  Creates a JDOM anchor element.  Can be overridden to change the URL creation,
     *  if you really know what you are doing.
     *  
     *  @param type One of the types above
     *  @param link URL to which to link to
     *  @param text Link text
     *  @param section If a particular section identifier is required.
     *  @return An A element.
     *  @since 2.4.78
     */
    protected Element createAnchor(Document doc, int type, String link, String text, String section) 
    { 
        Element el = doc.createElement("a");
        el.setAttribute("href",link+section); 
        addContent(el,text); 
        return el; 
    } 
    
    private Element makeLink(Document doc, int type, String link, String text, String section )
    {
        Element el = null;
        
        if( text == null ) text = link;

        text = callMutatorChain( m_linkMutators, text );

        section = (section != null) ? ("#"+section) : "";

        // Make sure we make a link name that can be accepted
        // as a valid URL.

        if( link.length() == 0 )
        {
            type = EMPTY;
        }

        switch(type)
        {
            case READ:
                el = createAnchor(doc, READ, link, text, section );
                break;

            case EDIT:
                el = createAnchor(doc, EDIT, link, text, "" ); 
                el.setAttribute("title","Create '"+link+"'");
                break;

            case EMPTY:
                el = doc.createElement("u");
                addContent(el,text);
                break;

                //
                //  These two are for local references - footnotes and 
                //  references to footnotes.
                //  We embed the page name (or whatever WikiContext gives us)
                //  to make sure the links are unique across Wiki.
                //
            case LOCALREF:
                //el = createAnchor( LOCALREF, "#ref-"+m_context.getName()+"-"+link, "["+text+"]", "" );
                //el = createAnchor( READ, link, text, section );
                break;

            case LOCAL:
               // el = new Element("a").setAttribute("class","footnote");
                //el.setAttribute("name", "ref-"+m_context.getName()+"-"+link.substring(1));
                //el.addContent("["+text+"]");
                break;

                //
                //  With the image, external and interwiki types we need to
                //  make sure nobody can put in Javascript or something else
                //  annoying into the links themselves.  We do this by preventing
                //  a haxor from stopping the link name short with quotes in 
                //  fillBuffer().
                //
            case IMAGE:
                el = doc.createElement("img");
                el.setAttribute("src",link);
                el.setAttribute("alt",text);
                break;

            case IMAGELINK:
                el = doc.createElement("img");
                el.setAttribute("src",link);
                el.setAttribute("alt",text);
                addContent(createAnchor(doc,IMAGELINK,text,"",""),el);
                break;

            case IMAGEWIKILINK:
                el = doc.createElement("img");
                el.setAttribute("src",link);
                el.setAttribute("alt",text);
                addContent(createAnchor(doc,IMAGEWIKILINK,link,"",""),el);
                break;

            case EXTERNAL:
                el = createAnchor(doc, EXTERNAL, link, text, section );
                if( m_useRelNofollow ) el.setAttribute("rel","nofollow");
                break;
                
            case INTERWIKI:
                el = createAnchor(doc, INTERWIKI, link, text, section );
                break;

            case ATTACHMENT:
                el = createAnchor(doc, ATTACHMENT, link, text, section );
                /*
                String attlink = m_context.getURL( WikiContext.ATTACH,
                                                   link );

                String infolink = m_context.getURL( WikiContext.INFO,
                                                    link );

                String imglink = m_context.getURL( WikiContext.NONE,
                                                   "images/attachment_small.png" );

                el = createAnchor( ATTACHMENT, attlink, text, "" );
                
                pushElement(el);
                popElement(el.getName());
             
                if( m_useAttachmentImage )
                {
                    el = new Element("img").setAttribute("src",imglink);
                    el.setAttribute("border","0");
                    el.setAttribute("alt","(info)");

                    el = new Element("a").setAttribute("href",infolink).addContent(el);
                }
                else
                {
                    el = null;
                }*/
                break;

            default:
                break;
        }

        if( el != null )
        {
            flushPlainText(doc);
            addContent(m_currentElement, el );
        }
        return el;
    }


    /**
     * 
     * copied from Arrays
     * 
     * Searches the specified array for the specified object using the binary
     * search algorithm.  The array must be sorted into ascending order
     * according to the <i>natural ordering</i> of its elements (as by
     * <tt>Sort(Object[]</tt>), above) prior to making this call.  If it is
     * not sorted, the results are undefined.
     * (If the array contains elements that are not  mutually comparable (for
     * example,strings and integers), it <i>cannot</i> be sorted according 
     * to the natural order of its elements, hence results are undefined.)
     *  If the array contains multiple
     * elements equal to the specified object, there is no guarantee which
     * one will be found.
     *
     * @param a the array to be searched.
     * @param key the value to be searched for.
     * @return index of the search key, if it is contained in the list;
     *	       otherwise, <tt>(-(<i>insertion point</i>) - 1)</tt>.  The
     *	       <i>insertion point</i> is defined as the point at which the
     *	       key would be inserted into the list: the index of the first
     *	       element greater than the key, or <tt>list.size()</tt>, if all
     *	       elements in the list are less than the specified key.  Note
     *	       that this guarantees that the return value will be &gt;= 0 if
     *	       and only if the key is found.
     * @throws ClassCastException if the search key in not comparable to the
     *         elements of the array.
     * @see Comparable
     * @see #sort(Object[])
     */
    public static int arrayBinarySearch(Object[] a, Object key) {
		int low = 0;
		int high = a.length-1;
	
		while (low <= high) {
		    int mid = (low + high) >> 1;
		    Comparable midVal = (Comparable)a[mid];
		    int cmp = midVal.compareTo(key);
	
		    if (cmp < 0)
			low = mid + 1;
		    else if (cmp > 0)
			high = mid - 1;
		    else
			return mid; // key found
		}
		return -(low + 1);  // key not found.
    }
    public static int  arrayBinarySearch(Object[] a, Object key, Comparator c) {
	        if (c==null) {
	            return arrayBinarySearch(a, key);
		}
	
		int low = 0;
		int high = a.length-1;
	
		while (low <= high) {
		    int mid = (low + high) >> 1;
		    Object midVal = a[mid];
		    int cmp = c.compare(midVal, key);
	
		    if (cmp < 0)
			low = mid + 1;
		    else if (cmp > 0)
			high = mid - 1;
		    else
			return mid; // key found
		}
		return -(low + 1);  // key not found.
    }
    
    /**
     *  Figures out if a link is an off-site link.  This recognizes
     *  the most common protocols by checking how it starts.
     *  
     *  @since 2.4
     */

    public static boolean isExternalLink( String link )
    {
        int idx = arrayBinarySearch( c_externalLinks, link, 
                                       c_startingComparator );

        //
        //  We need to check here once again; otherwise we might
        //  get a match for something like "h".
        //
        if( idx >= 0 && link.startsWith(c_externalLinks[idx]) ) return true;
        
        return false;
    }

    /**
     *  Returns true, if the link in question is an access
     *  rule.
     */
    private static boolean isAccessRule( String link )
    {
        return link.startsWith("{ALLOW") || link.startsWith("{DENY");
    }

    /**
     *  Matches the given link to the list of image name patterns
     *  to determine whether it should be treated as an inline image
     *  or not.
     */
    private boolean isImageLink( String link )
    {
        if( m_inlineImages )
        {
            for( Iterator i = m_inlineImagePatterns.iterator(); i.hasNext(); )
            {
                if( m_inlineMatcher.matches( link, (Pattern) i.next() ) )
                    return true;
            }
        }

        return false;
    }

    private static boolean isMetadata( String link )
    {
        return link.startsWith("{SET");
    }

    /**
     *  This method peeks ahead in the stream until EOL and returns the result.
     *  It will keep the buffers untouched.
     *
     *  @return The string from the current position to the end of line.
     */

    // FIXME: Always returns an empty line, even if the stream is full.
    private String peekAheadLine()
    {
        String s = readUntilEOL().toString();
        
        if( s.length() > PUSHBACK_BUFFER_SIZE )
        {
            //log.warn("Line is longer than maximum allowed size ("+PUSHBACK_BUFFER_SIZE+" characters.  Attempting to recover...");
            pushBack( s.substring(0,PUSHBACK_BUFFER_SIZE-1) );
        }
        else
        {
        	pushBack( s );
        }
        return s;
    }
    

    /**
     *  Writes HTML for error message.
     */

    public static Element makeError(Document doc, String error )
    {
    	Element el = doc.createElement("error");
    	el.appendChild(doc.createTextNode(error));
        return el;
        
    }

    private int flushPlainText(Document doc)
    {
        int numChars = m_plainTextBuf.length();
        
        if( numChars > 0 )
        {
            String buf;
            
            if( !m_allowHTML )
            {
                buf = escapeHTMLEntities(m_plainTextBuf);
            }
            else
            {
                buf = m_plainTextBuf.toString();
            }
            //
            //  We must first empty the buffer because the side effect of 
            //  calling makeCamelCaseLink() is to call this routine.
            //

            m_plainTextBuf = new StringBuffer(20);
             
            try
            {
                //
                //  This is the heaviest part of parsing, and therefore we can
                //  do some optimization here.
                //
                //  1) Only when the length of the buffer is big enough, we try to do the match
                //

                if( m_camelCaseLinks && !m_isEscaping && buf.length() > 3 )
                {            
                    // System.out.println("Buffer="+buf);

                    while( m_camelCaseMatcher.contains( buf, m_camelCasePattern ) )
                    {
                        MatchResult result = m_camelCaseMatcher.getMatch();

                        String firstPart = buf.substring(0,result.beginOffset(0)); 
                        String prefix = result.group(1);
                    
                        if( prefix == null ) prefix = "";
                    
                        String camelCase = result.group(2);
                        String protocol  = result.group(3);
                        String uri       = protocol+result.group(4);
                        buf              = buf.substring(result.endOffset(0));
                    
                        addContent(m_currentElement, firstPart );
                    
                        //
                        //  Check if the user does not wish to do URL or WikiWord expansion
                        //
                        if( prefix.endsWith("~") || prefix.indexOf('[') != -1 )
                        {
                            if( prefix.endsWith("~") ) prefix = prefix.substring(0,prefix.length()-1);
                            if( camelCase != null )
                            {
                                addContent( m_currentElement, prefix+camelCase );
                            }
                            else if( protocol != null )
                            {
                                addContent(m_currentElement, prefix+uri );
                            }
                            continue;
                        }
                    
                        //
                        //  Fine, then let's check what kind of a link this was
                        //  and emit the proper elements
                        //
                        if( protocol != null )
                        {
                            char c = uri.charAt(uri.length()-1);
                            if( c == '.' || c == ',' )
                            {
                                uri = uri.substring(0,uri.length()-1);
                                buf = c + buf;
                            }
                            // System.out.println("URI match "+uri);
                            addContent(m_currentElement, prefix );
                            makeDirectURILink(doc, uri );
                        }
                        else
                        {
                            // System.out.println("Matched: '"+camelCase+"'");
                            // System.out.println("Split to '"+firstPart+"', and '"+buf+"'");
                            // System.out.println("prefix="+prefix);
                            addContent(m_currentElement, prefix );
                        
                            makeCamelCaseLink(doc, camelCase );
                        }
                    }

                    addContent(m_currentElement, buf );
                }
                else
                {
                    //
                    //  No camelcase asked for, just add the elements
                    //
                   addContent( m_currentElement, buf );
                }
            }
            catch( Exception e )
            {
                //
                // Sometimes it's possible that illegal XML chars is added to the data.
                // Here we make sure it does not stop parsing.
                //
                addContent(m_currentElement, makeError(m_currentElement.getOwnerDocument(), e.getMessage()) );
            }
        }
        
        return numChars;
    }

    private void addContent(Element element, Element element2) {
		element.appendChild(element2);
	}

	private void addContent(Element element, String text) {

		if (text == null) return;
		
		//avoid empty lines
		if ("".equals(text.trim())) return;
		element.appendChild(element.getOwnerDocument().createTextNode(text));
		
	}

	/**
     *  Escapes XML entities in a HTML-compatible way (i.e. does not escape
     *  entities that are already escaped).
     *  
     *  @param buf
     *  @return
     */
    private String escapeHTMLEntities(StringBuffer buf)
    {
        StringBuffer tmpBuf = new StringBuffer( buf.length() + 20 );
        
        for( int i = 0; i < buf.length(); i++ )
        {
            char ch = buf.charAt(i);
            
            if( ch == '<' ) 
            {
                tmpBuf.append("&lt;");
            }
            else if( ch == '>' ) 
            {
                tmpBuf.append("&gt;");
            }
            else if( ch == '&' )
            {
                for( int j = (i < buf.length()-1 ) ? i+1 : i; j < buf.length(); j++ )
                {
                    int ch2 = buf.charAt(j);
                    if( ch2 == ';' ) 
                    { 
                        tmpBuf.append(ch); 
                        break; 
                    }
                    if( ch2 != '#' && !Character.isLetterOrDigit( (char)ch2) ) 
                    { 
                        tmpBuf.append("&amp;"); break; 
                    }
                }
            }
            else
            {
                tmpBuf.append( ch );
            }
        }

        return tmpBuf.toString();
    }
    
    private Element pushElement( Element e )
    {
        flushPlainText(m_currentElement.getOwnerDocument());
        addContent(m_currentElement, e );
        m_currentElement = e;
        
        return e;
    }
    
    private Element addElement( Node e )
    {
        if( e != null )
        {
            flushPlainText(m_currentElement.getOwnerDocument());
            m_currentElement.appendChild( e );
        }
        return m_currentElement;
    }
    
    /**
     *  All elements that can be empty by the HTML DTD.
     */
    //  Keep sorted.
    private static final String[] EMPTY_ELEMENTS = {
        "area", "base", "br", "col", "hr", "img", "input", "link", "meta", "p", "param"
    };
    
    private Element popElement( String s )
    {
    	Document doc = m_currentElement.getOwnerDocument();
        int flushedBytes = flushPlainText(doc);
    
        Node currEl = m_currentElement;
        
        while( currEl.getParentNode() != null )
        {
            if( (currEl instanceof Element) && ((Element) currEl).getTagName().equals(s)&& (currEl.getOwnerDocument().getDocumentElement() != currEl) )
            {
                Node node = currEl.getParentNode();
                if ( node instanceof Element) {
	                m_currentElement = (Element) node;
	                
	                if( flushedBytes == 0 && arrayBinarySearch( EMPTY_ELEMENTS, s ) < 0 )
	                {
	                	/*if (currEl.hasChildNodes()) {
	                		NodeList nl = currEl.getChildNodes();
	                		for (int i = 0; i < nl.getLength(); i++) {
	                			Node child = nl.item(i);
	                			if (child instanceof Text) 
	                				((Text) child).setData("");
	                		}
	                	}*/
	                }
	                
	                return m_currentElement;
                }
            }
            
            currEl = currEl.getParentNode();
        }
        
        return m_currentElement;
    }


    /**
     *  Reads the stream until it meets one of the specified
     *  ending characters, or stream end.  The ending character will be left
     *  in the stream.
     */
    private String readUntil( String endChars )
    {
        StringBuffer sb = new StringBuffer( 80 );
        int ch = nextToken();

        while( ch != -1 )
        {
            if( ch == '\\' ) 
            {
                ch = nextToken(); 
                if( ch == -1 ) 
                {
                    break;
                }
            }
            else
            {
                if( endChars.indexOf((char)ch) != -1 )
                {
                    pushBack( ch );
                    break;
                }
            }
            sb.append( (char) ch );
            ch = nextToken();
        }

        return sb.toString();
    }

    /**
     *  Reads the stream while the characters that have been specified are
     *  in the stream, returning then the result as a String.
     */
    private String readWhile( String endChars )
    {
        StringBuffer sb = new StringBuffer( 80 );
        int ch = nextToken();

        while( ch != -1 )
        {
            if( endChars.indexOf((char)ch) == -1 )
            {
                pushBack( ch );
                break;
            }
            
            sb.append( (char) ch );
            ch = nextToken();
        }

        return sb.toString();
    }

    private JSPWikiMarkupParser m_cleanTranslator;

    /**
     *  Does a lazy init.  Otherwise, we would get into a situation
     *  where HTMLRenderer would try and boot a TranslatorReader before
     *  the TranslatorReader it is contained by is up.
     */
    private JSPWikiMarkupParser getCleanTranslator()
    {
        if( m_cleanTranslator == null )
        {        
            m_cleanTranslator = new JSPWikiMarkupParser( null );

            m_cleanTranslator.m_allowHTML = true;
        }

        return m_cleanTranslator;
    }
    /**
     *  Modifies the "hd" parameter to contain proper values.  Because
     *  an "id" tag may only contain [a-zA-Z0-9:_-], we'll replace the
     *  % after url encoding with '_'.
     */
    /*// FIXME: This method should probably be public and in an util class somewhere
    private String makeHeadingAnchor( String baseName, String title, Heading hd )
    {
        hd.m_titleText = title;
        title = MarkupParser.cleanLink( title );
        hd.m_titleSection = m_engine.encodeName(title);
        hd.m_titleAnchor = "section-"+m_engine.encodeName(baseName)+
                           "-"+hd.m_titleSection;
        
        hd.m_titleAnchor = hd.m_titleAnchor.replace( '%', '_' );
        hd.m_titleAnchor = hd.m_titleAnchor.replace( '/', '_' );
        return hd.m_titleAnchor;
    }*/

    private String makeSectionTitle( String title )
    {
        title = title.trim();
        String outTitle = title;
        
     /*   try
        {
            JSPWikiMarkupParser dtr = getCleanTranslator();
            dtr.setInputReader( new StringReader(title) );

            CleanTextRenderer ctt = new CleanTextRenderer(m_context, dtr.parse());
            
            outTitle = ctt.getString();
        }
        catch( IOException e )
        {
            //log.fatal("CleanTranslator not working", e);
            throw new RuntimeException("CleanTranslator not working as expected, when cleaning title"+ e.getMessage() );
        }*/

        return outTitle;
    }
    
    /**
     *  Returns XHTML for the start of the heading.  Also sets the
     *  line-end emitter.
     *  @param level
     *  @param title the title for the heading
     *  @param hd a List to which heading should be added
     */ 
    public Element makeHeading(Document doc, int level, String title, Heading hd )
    {
        Element el = null;
        
        String outTitle = makeSectionTitle( title );

        hd.m_level = level;

        switch( level )
        {
          case Heading.HEADING_SMALL:
            el = doc.createElement("h4");
            break;

          case Heading.HEADING_MEDIUM:
            el = doc.createElement("h3");
            break;

          case Heading.HEADING_LARGE:
            el = doc.createElement("h2");
            break;
        }
        
        return el;
    }

    /**
     *  When given a link to a WikiName, we just return
     *  a proper HTML link for it.  The local link mutator
     *  chain is also called.
     */
    private Element makeCamelCaseLink(Document doc, String wikiname )
    {
        String matchedLink;

        callMutatorChain( m_localLinkMutatorChain, wikiname );

       /* if( (matchedLink = linkExists( wikiname )) != null )
        {
            makeLink( READ, matchedLink, wikiname, null );
        }
        else
        {*/
            makeLink(doc, EDIT, wikiname, wikiname, null );
       // }

        return m_currentElement;
    }

    /** Holds the image URL for the duration of this parser */
    private String m_outlinkImageURL = null;
    
    /**
     *  Returns an element for the external link image (out.png).  However,
     *  this method caches the URL for the lifetime of this MarkupParser,
     *  because it's commonly used, and we'll end up with possibly hundreds
     *  our thousands of references to it...  It's a lot faster, too.
     *  
     *  @return  An element containing the HTML for the outlink image.
     */
    private Element outlinkImage(Document doc)
    {
        Element el = null;
        
      /*  if( m_useOutlinkImage )
        {
            if( m_outlinkImageURL == null )
            {
                m_outlinkImageURL = m_context.getURL( WikiContext.NONE, OUTLINK_IMAGE );
            }
            
            el = new Element("img").setAttribute("class", "outlink");
            el.setAttribute( "src", m_outlinkImageURL );
            el.setAttribute("alt","");
        }*/

        return el;
    }

    /**
     *  Takes an URL and turns it into a regular wiki link.  Unfortunately,
     *  because of the way that flushPlainText() works, it already encodes
     *  all of the XML entities.  But so does WikiContext.getURL(), so we
     *  have to do a reverse-replace here, so that it can again be replaced in makeLink.
     *  <p>
     *  What a crappy problem.
     *  
     * @param url
     * @return
     */
    private Element makeDirectURILink(Document doc, String url )
    {
        Element result;
        String last = null;
        
        if( url.endsWith(",") || url.endsWith(".") )
        {
            last = url.substring( url.length()-1 );
            url  = url.substring( 0, url.length()-1 );
        }

        callMutatorChain( m_externalLinkMutatorChain, url );

        if( isImageLink( url ) )
        {
            result = handleImageLink(doc, StringUtils.replace(url,"&amp;","&"), url, false );
        }
        else
        {
            result = makeLink(doc, EXTERNAL, StringUtils.replace(url,"&amp;","&"), url, null );
            addElement( outlinkImage(doc) );
        }

        if( last != null )
        {
            m_plainTextBuf.append(last);
        }
        
        return result;
    }

    // A simplified copy of org.java.commons.lang.StringUtils
    public static class StringUtils {
	    public static boolean isEmpty(String str) {
	        return str == null || str.length() == 0;
	    }
	    public static String replace(String text, String repl, String with) {
	        return replace(text, repl, with, -1);
	    }
	    
	    public static String replace(String text, String repl, String with, int max) {
	        if (StringUtils.isEmpty(text) || isEmpty(repl) || with == null || max == 0) {
	            return text;
	        }
	        int start = 0;
	        int end = text.indexOf(repl, start);
	        if (end == -1) {
	            return text;
	        }
	        int replLength = repl.length();
	        int increase = with.length() - replLength;
	        increase = (increase < 0 ? 0 : increase);
	        increase *= (max < 0 ? 16 : (max > 64 ? 64 : max));
	        String buf = "";
	        while (end != -1) {
	            buf = buf + text.substring(start, end) + with;
	            start = end + replLength;
	            if (--max == 0) {
	                break;
	            }
	            end = text.indexOf(repl, start);
	        }
	        buf += text.substring(start);
	        return buf;
	    }
    }

    /**
     *  Image links are handled differently:
     *  1. If the text is a WikiName of an existing page,
     *     it gets linked.
     *  2. If the text is an external link, then it is inlined.  
     *  3. Otherwise it becomes an ALT text.
     *
     *  @param reallink The link to the image.
     *  @param link     Link text portion, may be a link to somewhere else.
     *  @param hasLinkText If true, then the defined link had a link text available.
     *                  This means that the link text may be a link to a wiki page,
     *                  or an external resource.
     */
    
    // FIXME: isExternalLink() is called twice.
    private Element handleImageLink(Document doc, String reallink, String link, boolean hasLinkText )
    {
        String possiblePage = MarkupParser.cleanLink( link );

        if( isExternalLink( link ) && hasLinkText )
        {
            return makeLink(doc, IMAGELINK, reallink, link, null );
        }
      /*  else if( ( linkExists( possiblePage ) ) != null &&
                 hasLinkText )
        {
            // System.out.println("Orig="+link+", Matched: "+matchedLink);
            callMutatorChain( m_localLinkMutatorChain, possiblePage );
            
            return makeLink( IMAGEWIKILINK, reallink, link, null );
        }*/
        else
        {
            return makeLink(doc, IMAGE, reallink, link, null );
        }
    }
    
    /** 
     *  Parses an access rule.
     *
     *  Note that this method
     *  manages the realPage instead of the page declared in the context. This
     *  ensures that variable definitions do not leak when included when InsertPageTag.
     *  
     */
    private Element handleAccessRule( String ruleLine )
    {
        /*if( !m_parseAccessRules ) return m_currentElement;
        Acl acl;
        WikiPage          page = m_context.getRealPage();
        // UserDatabase      db = m_context.getEngine().getUserDatabase();

        if( ruleLine.startsWith( "{" ) )
            ruleLine = ruleLine.substring( 1 );
        if( ruleLine.endsWith( "}" ) )
            ruleLine = ruleLine.substring( 0, ruleLine.length() - 1 );

        log.debug("page="+page.getName()+", ACL = "+ruleLine);
        
        try
        {
            acl = m_engine.getAclManager().parseAcl( page, ruleLine );

            page.setAcl( acl );

            log.debug( acl.toString() );
        }
        catch( WikiSecurityException wse )
        {
            return makeError( wse.getMessage() );
        }*/

        return m_currentElement;
    }

    /**
     *  Handles metadata setting [{SET foo=bar}].  Note that this method
     *  handles the realPage instead of the page declared in the context. This
     *  ensures that variable definitions do not leak when included when InsertPageTag.
     */
    private Element handleMetadata( String link )
    {
      /*  try
        {
            String args = link.substring( link.indexOf(' '), link.length()-1 );
            
            String name = args.substring( 0, args.indexOf('=') );
            String val  = args.substring( args.indexOf('=')+1, args.length() );

            name = name.trim();
            val  = val.trim();

            if( val.startsWith("'") ) val = val.substring( 1 );
            if( val.endsWith("'") )   val = val.substring( 0, val.length()-1 );

            // log.debug("SET name='"+name+"', value='"+val+"'.");

            if( name.length() > 0 && val.length() > 0 )
            {
                val = m_engine.getVariableManager().expandVariables( m_context,
                                                                     val );
            
                m_context.getRealPage().setAttribute( name, val );
            }
        }
        catch( Exception e )
        {
            return makeError(" Invalid SET found: "+link);
        }*/

        return m_currentElement;
    }


    /**
     *  Emits a processing instruction that will disable markup escaping. This is
     *  very useful if you want to emit HTML directly into the stream.
     *
     */
    private void disableOutputEscaping(Document doc)
    {
        addElement( doc.createProcessingInstruction(PI_DISABLE_OUTPUT_ESCAPING, "") );
    }
    
    /**
     *  Gobbles up all hyperlinks that are encased in square brackets.
     */
    private Element handleHyperlinks(Document doc, String link, int pos )
    {
        StringBuffer sb        = new StringBuffer(link.length()+80);
        String       reallink;
        int          cutpoint;

        if( isAccessRule( link ) )
        {
            return handleAccessRule( link );
        }

        if( isMetadata( link ) )
        {
            return handleMetadata( link );
        }

        /*if( PluginManager.isPluginLink( link ) )
        {
            try
            {
                Content pluginContent = m_engine.getPluginManager().parsePluginLine( m_context, link, pos );

                addElement( pluginContent );
            }
            catch( PluginException e )
            {
                log.info( "Failed to insert plugin", e );
                log.info( "Root cause:",e.getRootThrowable() );
                return addElement( makeError("Plugin insertion failed: "+e.getMessage()) );
            }
            
            return m_currentElement;
        }*/

        // link = TextUtil.replaceEntities( link );

        if( (cutpoint = link.indexOf('|')) != -1 )
        {                    
            reallink = link.substring( cutpoint+1 ).trim();
            link = link.substring( 0, cutpoint );
        }
        else
        {
            reallink = link.trim();
        }

        int interwikipoint = -1;

        //
        //  Yes, we now have the components separated.
        //  link     = the text the link should have
        //  reallink = the url or page name.
        //
        //  In many cases these are the same.  [link|reallink].
        //  
      /*  if( VariableManager.isVariableLink( link ) )
        {
            Content el = new VariableContent(link);

            addElement( el );
        }
        else if( isExternalLink( reallink ) )
        {
            // It's an external link, out of this Wiki

            callMutatorChain( m_externalLinkMutatorChain, reallink );

            if( isImageLink( reallink ) )
            {
                handleImageLink( reallink, link, (cutpoint != -1) );
            }
            else
            {
                makeLink( EXTERNAL, reallink, link, null );
                addElement( outlinkImage() );
            }
        }
        else if( (interwikipoint = reallink.indexOf(":")) != -1 )
        {
            // It's an interwiki link
            // InterWiki links also get added to external link chain
            // after the links have been resolved.
            
            // FIXME: There is an interesting issue here:  We probably should
            //        URLEncode the wikiPage, but we can't since some of the
            //        Wikis use slashes (/), which won't survive URLEncoding.
            //        Besides, we don't know which character set the other Wiki
            //        is using, so you'll have to write the entire name as it appears
            //        in the URL.  Bugger.
            
            String extWiki = reallink.substring( 0, interwikipoint );
            String wikiPage = reallink.substring( interwikipoint+1 );

            String urlReference = m_engine.getInterWikiURL( extWiki );

            if( urlReference != null )
            {
                urlReference = TextUtil.replaceString( urlReference, "%s", wikiPage );
                urlReference = callMutatorChain( m_externalLinkMutatorChain, urlReference );

                if( isImageLink(urlReference) )
                {
                    handleImageLink( urlReference, link, cutpoint != -1 );
                }
                else
                {
                    makeLink( INTERWIKI, urlReference, link, null );
                }
                
                if( isExternalLink(urlReference) )
                {
                    addElement( outlinkImage() );
                }
            }
            else
            {
                addElement( makeError("No InterWiki reference defined in properties for Wiki called '"+extWiki+"'!)") );
            }
        }
        else if( reallink.startsWith("#") )
        {
            // It defines a local footnote
            makeLink( LOCAL, reallink, link, null );
        }
        else if( TextUtil.isNumber( reallink ) )
        {
            // It defines a reference to a local footnote
            makeLink( LOCALREF, reallink, link, null );
        }
        else
        {
            int hashMark = -1;

            //
            //  Internal wiki link, but is it an attachment link?
            //
            String attachment = findAttachment( reallink );
            if( attachment != null )
            {
                callMutatorChain( m_attachmentLinkMutatorChain, attachment );

                if( isImageLink( reallink ) )
                {
                    attachment = m_context.getURL( WikiContext.ATTACH, attachment );
                    sb.append( handleImageLink( attachment, link, (cutpoint != -1) ) );
                }
                else
                {
                    makeLink( ATTACHMENT, attachment, link, null );
                }
            }
            else if( (hashMark = reallink.indexOf('#')) != -1 )
            {
                // It's an internal Wiki link, but to a named section

                String namedSection = reallink.substring( hashMark+1 );
                reallink = reallink.substring( 0, hashMark );

                reallink     = MarkupParser.cleanLink( reallink );

                callMutatorChain( m_localLinkMutatorChain, reallink );

                String matchedLink;
                if( (matchedLink = linkExists( reallink )) != null )
                {
                    String sectref = "section-"+m_engine.encodeName(matchedLink)+"-"+namedSection;
                    sectref = sectref.replace('%', '_');
                    makeLink( READ, matchedLink, link, sectref );
                }
                else
                {
                    makeLink( EDIT, reallink, link, null );
                }
            }
            else
            {
                // It's an internal Wiki link
                reallink = MarkupParser.cleanLink( reallink );

                callMutatorChain( m_localLinkMutatorChain, reallink );

                String matchedLink = linkExists( reallink );
                
                if( matchedLink != null )
                {
                    makeLink( READ, matchedLink, link, null );
                }
                else
                {
                    makeLink( EDIT, reallink, link, null );
                }
            }
        } */
        makeLink(doc, READ, reallink, link, null );
        
        return m_currentElement;
    }

    private String findAttachment( String link )
    {
       /* AttachmentManager mgr = m_engine.getAttachmentManager();
        Attachment att = null;

        try
        {
            att = mgr.getAttachmentInfo( m_context, link );
        }
        catch( ProviderException e )
        {
            //log.warn("Finding attachments failed: ",e);
            return null;
        }

        if( att != null )
        {
            return att.getName();
        }
        else*/
    	if( link.indexOf('/') != -1 )
        {
            return link;
        }

        return null;
    }

    /**
     *  Pushes back any string that has been read.  It will obviously
     *  be pushed back in a reverse order.
     *
     *  @since 2.1.77
     */
    private void pushBack( String s )
    {
        for( int i = s.length()-1; i >= 0; i-- )
        {
            pushBack( s.charAt(i) );
        }
    }

    private Element handleBackslash(Document doc)
    {
        int ch = nextToken();

        if( ch == '\\' )
        {
            int ch2 = nextToken();

            if( ch2 == '\\' )
            {
                pushElement( doc.createElement("br"));
                return popElement("br");
            }
           
            pushBack( ch2 );

            pushElement( doc.createElement("br") );
            return popElement("br");
        }

        pushBack( ch );

        return null;
    }

    private Element handleUnderscore(Document doc)
    {
        int ch = nextToken();
        Element el = null;
        
        if( ch == '_' )
        {
            if( m_isbold )
            {
                el = popElement("b");
            }
            else
            {
                el = pushElement( doc.createElement("b") );
            }
            m_isbold = !m_isbold;
        }
        else
        {
            pushBack( ch );
        }

        return el;
    }


    /**
     *  For example: italics.
     */
    private Element handleApostrophe(Document doc)
    {
        int ch = nextToken();
        Element el = null;
        
        if( ch == '\'' )
        {
            if( m_isitalic )
            {
                el = popElement("i");
            }
            else
            {
                el = pushElement( doc.createElement("i") );
            }
            m_isitalic = !m_isitalic;
        }
        else
        {
            pushBack( ch );
        }

        return el;
    }

    private Element handleOpenbrace(Document doc, boolean isBlock )
    {
        int ch = nextToken();

        if( ch == '{' )
        {
            int ch2 = nextToken();

            if( ch2 == '{' )
            {
                m_isPre = true;
                m_isEscaping = true;
                m_isPreBlock = isBlock;
                
                if( isBlock )
                {
                    startBlockLevel();
                    return pushElement(doc.createElement("pre") );
                }
                
                Element el = doc.createElement("span");
                el.setAttribute("style","font-family:monospace; white-space:pre;");

                return pushElement( el );
            }

            pushBack( ch2 );
            
            return pushElement( doc.createElement("tt") );
        }
        
        pushBack( ch );

        return null;
    }

    /**
     *  Handles both }} and }}}
     */
    private Element handleClosebrace()
    {
        int ch2 = nextToken();

        if( ch2 == '}' )
        {
            int ch3 = nextToken();

            if( ch3 == '}' )
            {
                if( m_isPre )
                {
                    if( m_isPreBlock )
                    {
                        popElement( "pre" );
                    }
                    else
                    {
                        popElement( "span" );
                    }

                    m_isPre = false;
                    m_isEscaping = false;
                    return m_currentElement;
                }

                m_plainTextBuf.append("}}}");
                return m_currentElement;
            }

            pushBack( ch3 );

            if( !m_isEscaping )
            {
                return popElement("tt");
            }
        }

        pushBack( ch2 );

        return null;
    }

    private Element handleDash(Document doc)
    {
        int ch = nextToken();

        if( ch == '-' )
        {
            int ch2 = nextToken();

            if( ch2 == '-' )
            {
                int ch3 = nextToken();

                if( ch3 == '-' ) 
                {
                    // Empty away all the rest of the dashes.
                    // Do not forget to return the first non-match back.
                    while( (ch = nextToken()) == '-' );
                    
                    pushBack(ch);
                    startBlockLevel();
                    pushElement( doc.createElement("hr") );
                    return popElement( "hr" );
                }
        
                pushBack( ch3 );
            }
            pushBack( ch2 );
        }

        pushBack( ch );

        return null;
    }

    private Element handleHeading(Document doc)
    {
        Element el = null;
        
        int ch  = nextToken();
        if (ch < 0) return null;

        Heading hd = new Heading();

        if( ch == '!' )
        {
            int ch2 = nextToken();

            if( ch2 == '!' )
            {
                String title = peekAheadLine();
                
                el = makeHeading(doc, Heading.HEADING_LARGE, title, hd);
            }
            else
            {
                pushBack( ch2 );
                String title = peekAheadLine();
                el = makeHeading(doc, Heading.HEADING_MEDIUM, title, hd );
            }
        }
        else
        {
            pushBack( ch );
            String title = peekAheadLine();
            el = makeHeading(doc, Heading.HEADING_SMALL, title, hd );
        }

        callHeadingListenerChain( hd );

        if( el != null ) pushElement(el);
        
        return el;
    }

    /**
     *  Reads the stream until the next EOL or EOF.  Note that it will also read the
     *  EOL from the stream.
     */
    private StringBuffer readUntilEOL()
    {
        int ch;
        StringBuffer buf = new StringBuffer( 256 );

        while( true )
        {
            ch = nextToken();
            if (ch < 0)break;

            if( ch == -1 )
                break;

            buf.append( (char) ch );

            if( ch == '\n' ) 
                break;
        }
        return buf;
    }

    /** Controls whether italic is restarted after a paragraph shift */
    
    private boolean m_restartitalic = false;
    private boolean m_restartbold   = false;
    
    /**
     *  Starts a block level element, therefore closing
     *  a potential open paragraph tag.
     */
    private void startBlockLevel()
    {
        // These may not continue over block level limits in XHTML
        
        popElement("i");
        popElement("b");
        popElement("tt");

        if( m_isOpenParagraph )
        {
            m_isOpenParagraph = false;
            popElement("p");
            m_plainTextBuf.append("\n"); // Just small beautification
        }

        m_restartitalic = m_isitalic;
        m_restartbold   = m_isbold;
        
        m_isitalic = false;
        m_isbold   = false;
    }

    private static String getListType( char c )
    {
        if( c == '*' )
        {
            return "ul";
        }
        else if( c == '#' )
        {
            return "ol";
        }
        throw new RuntimeException("Parser got faulty list type: "+c);
    }
    /**
     *  Like original handleOrderedList() and handleUnorderedList()
     *  however handles both ordered ('#') and unordered ('*') mixed together.
     */

    // FIXME: Refactor this; it's a bit messy.

    private Element handleGeneralList(Document doc)
    {
         startBlockLevel();

         String strBullets = readWhile( "*#" );
         // String strBulletsRaw = strBullets;      // to know what was original before phpwiki style substitution
         int numBullets = strBullets.length();

         // override the beginning portion of bullet pattern to be like the previous
         // to simulate PHPWiki style lists

         if(m_allowPHPWikiStyleLists)
         {
             // only substitute if different
             if(!( strBullets.substring(0,Math.min(numBullets,m_genlistlevel)).equals
                   (m_genlistBulletBuffer.substring(0,Math.min(numBullets,m_genlistlevel)) ) ) )
             {
                 if(numBullets <= m_genlistlevel)
                 {
                     // Substitute all but the last character (keep the expressed bullet preference)
                     strBullets  = (numBullets > 1 ? m_genlistBulletBuffer.substring(0, numBullets-1) : "")
                                   + strBullets.substring(numBullets-1, numBullets);
                 }
                 else
                 {
                     strBullets = m_genlistBulletBuffer + strBullets.substring(m_genlistlevel, numBullets);
                 }
             }
         }

         //
         //  Check if this is still of the same type
         //
         if( strBullets.substring(0,Math.min(numBullets,m_genlistlevel)).equals
            (m_genlistBulletBuffer.substring(0,Math.min(numBullets,m_genlistlevel)) ) )
         {
             if( numBullets > m_genlistlevel )
             {
                 pushElement( doc.createElement( getListType(strBullets.charAt(m_genlistlevel++) ) ) );
                 // buf.append( m_renderer.openList(strBullets.charAt(m_genlistlevel++)) );

                 for( ; m_genlistlevel < numBullets; m_genlistlevel++ )
                 {
                     // bullets are growing, get from new bullet list
                     pushElement( doc.createElement("li") );
                     // buf.append( m_renderer.openListItem() );
                     pushElement( doc.createElement( getListType(strBullets.charAt(m_genlistlevel)) ));
                     // buf.append( m_renderer.openList(strBullets.charAt(m_genlistlevel)) );
                 }
             }
             else if( numBullets < m_genlistlevel )
             {
                 //  Close the previous list item.
                 // buf.append( m_renderer.closeListItem() );
                 popElement( "li" );
                 
                 for( ; m_genlistlevel > numBullets; m_genlistlevel-- )
                 {
                     // bullets are shrinking, get from old bullet list
                     // buf.append( m_renderer.closeList(m_genlistBulletBuffer.charAt(m_genlistlevel - 1)) );
                     
                     popElement( getListType(m_genlistBulletBuffer.charAt(m_genlistlevel-1)) );
                     if( m_genlistlevel > 0 ) 
                     {
                         // buf.append( m_renderer.closeListItem() );
                         popElement( "li" );
                     }

                 }
             }
             else
             {
                 if( m_genlistlevel > 0 ) 
                 {
                     popElement( "li" );
                     // buf.append( m_renderer.closeListItem() );
                 }
             }
         }
         else
         {
             //
             //  The pattern has changed, unwind and restart
             //
             int  numEqualBullets;
             int  numCheckBullets;

             // find out how much is the same
             numEqualBullets = 0;
             numCheckBullets = Math.min(numBullets,m_genlistlevel);

             while( numEqualBullets < numCheckBullets )
             {
                 // if the bullets are equal so far, keep going
                 if( strBullets.charAt(numEqualBullets) == m_genlistBulletBuffer.charAt(numEqualBullets))
                     numEqualBullets++;
                 // otherwise giveup, we have found how many are equal
                 else
                     break;
             }

             //unwind
             for( ; m_genlistlevel > numEqualBullets; m_genlistlevel-- )
             {
                 popElement( getListType( m_genlistBulletBuffer.charAt(m_genlistlevel-1) ) );
                 // buf.append( m_renderer.closeList( m_genlistBulletBuffer.charAt(m_genlistlevel - 1) ) );
                 if( m_genlistlevel > 0 ) 
                 {
                     //buf.append( m_renderer.closeListItem() );
                     popElement("li");
                 }
             }

             //rewind
             // buf.append( m_renderer.openList( strBullets.charAt(numEqualBullets++) ) );
             pushElement( doc.createElement(getListType( strBullets.charAt(numEqualBullets++) ) ) );
             for(int i = numEqualBullets; i < numBullets; i++)
             {
                 pushElement( doc.createElement("li") );
                 pushElement( doc.createElement( getListType( strBullets.charAt(i) ) ) );
                 // buf.append( m_renderer.openListItem() );
                 // buf.append( m_renderer.openList( strBullets.charAt(i) ) );
             }
             m_genlistlevel = numBullets;
         }
         //buf.append( m_renderer.openListItem() );
         pushElement( doc.createElement("li") );
         
         // work done, remember the new bullet list (in place of old one)
         m_genlistBulletBuffer.setLength(0);
         m_genlistBulletBuffer.append(strBullets);

         return m_currentElement;
    }

    private Element unwindGeneralList()
    {
        //unwind
        for( ; m_genlistlevel > 0; m_genlistlevel-- )
        {
            popElement( "li" );
            popElement( getListType(m_genlistBulletBuffer.charAt(m_genlistlevel-1)) );
        }

        m_genlistBulletBuffer.setLength(0);

        return null;
    }


    private Element handleDefinitionList(Document doc)
    {
        if( !m_isdefinition )
        {
            m_isdefinition = true;

            startBlockLevel();
            
            pushElement( doc.createElement("dl") );
            return pushElement( doc.createElement("dt") ); 
        }

        return null;
    }

    private Element handleOpenbracket(Document doc)
    {
        StringBuffer sb = new StringBuffer(40);
        int pos = getPosition();
        int ch;
        boolean isPlugin = false;

        while( (ch = nextToken()) == '[' )
        {
            sb.append( (char)ch );
        }

        if( ch == '{' )
        {
            isPlugin = true;
        }

        pushBack( ch );

        if( sb.length() > 0 )
        {
            m_plainTextBuf.append( sb );
            return m_currentElement;
        }

        //
        //  Find end of hyperlink
        //

        ch = nextToken();
        if (ch < 0) return null;
        
        int nesting = 1;    // Check for nested plugins

        while( ch != -1 )
        {
            int ch2 = nextToken(); pushBack(ch2);
            
            if( isPlugin ) 
            {
                if( ch == '[' && ch2 == '{' ) 
                {
                    nesting++;
                }
                else if( nesting == 0 && ch == ']' && sb.charAt(sb.length()-1) == '}' ) 
                {
                    break;
                }
                else if( ch == '}' && ch2 == ']' ) 
                {
                    // NB: This will be decremented once at the end
                    nesting--;
                }
            } 
            else 
            {
                if( ch == ']' ) 
                {
                    break;
                }
            }

            sb.append( (char) ch );

            ch = nextToken();
        }

        //
        //  If the link is never finished, do some tricks to display the rest of the line
        //  unchanged.
        //
        if( ch == -1 )
        {
           // log.debug("Warning: unterminated link detected!");
            m_isEscaping = true;
            m_plainTextBuf.append( sb );
            flushPlainText(doc);
            m_isEscaping = false;
            return m_currentElement;
        }

        return handleHyperlinks(doc, sb.toString(), pos );
    }

    /**
     *  Reads the stream until the current brace is closed or stream end. 
     */
    private String readBraceContent( char opening, char closing )
    {
        StringBuffer sb = new StringBuffer(40);
        int braceLevel = 1;    
        int ch;        
        while(( ch = nextToken() ) != -1 )
        {
            if (ch < 0) break;
            if( ch == '\\' ) 
            {
                continue;
            }
            else if ( ch == opening ) 
            {
                braceLevel++;
            }
            else if ( ch == closing ) 
            {
                braceLevel--;
                if (braceLevel==0) 
                {
                  break;
                }
            }
            sb.append( (char)ch );
        }    
        return sb.toString();
    }

    
    /**
     *  Handles constructs of type %%(style) and %%class
     * @param newLine
     * @return
     * @throws IOException
     */
    private Element handleDiv(Document doc, boolean newLine )
    {
        int ch = nextToken();
        if (ch < 0) return null;
        Element el = null;
        
        if( ch == '%' )
        {
            String style = null;
            String clazz = null;

            ch = nextToken();
            
            //
            //  Style or class?
            //
            if( ch == '(' )
            {                
                style = readBraceContent('(',')');
            }
            else if( Character.isLetter( (char) ch ) )
            {
                pushBack( ch );
                clazz = readUntil( " \t\n\r" );
                ch = nextToken();
                
                //
                //  Pop out only spaces, so that the upcoming EOL check does not check the
                //  next line.
                //
                if( ch == '\n' || ch == '\r' )
                {
                    pushBack(ch);
                }
            }
            else
            {
                //
                // Anything else stops.
                //

                pushBack(ch);
                
                try
                {
                    Boolean isSpan = (Boolean)m_styleStack.pop();
                
                    if( isSpan == null )
                    {
                        // Fail quietly
                    }
                    else if( isSpan.booleanValue() )
                    {
                        el = popElement( "span" );
                    }
                    else
                    {
                        el = popElement( "div" );
                    }
                }
                catch( EmptyStackException e )
                {
                    //log.debug("Page '"+m_context.getName()+"' closes a %%-block that has not been opened.");
                    return m_currentElement;
                }
                
                return el;
            }

            //
            //  Check if there is an attempt to do something nasty
            //
           /* style = StringEscapeUtils.unescapeHtml(style);
            if( style != null && style.indexOf("javascript:") != -1 )
            {
                //log.debug("Attempt to output javascript within CSS:"+style);
                return addElement( makeError(doc, "Attempt to output javascript!") );
            }*/
            
            //
            //  Decide if we should open a div or a span?
            //
            String eol = peekAheadLine();
            
            if( eol.trim().length() > 0 )
            {
                // There is stuff after the class
                
                el = doc.createElement("span");
                
                m_styleStack.push( Boolean.TRUE );
            }
            else
            {
                startBlockLevel();
                el = doc.createElement("div");
                m_styleStack.push( Boolean.FALSE );
            }

            if( style != null ) el.setAttribute("style", style);
            if( clazz != null ) el.setAttribute("class", clazz );
            el = pushElement( el );
            
            return el;
        }

        pushBack(ch);

        return el;
    }

    private Element handleSlash(Document doc, boolean newLine ) 
    {
        int ch = nextToken();
        if (ch < 0) return null;
        
        pushBack(ch);
        if( ch == '%' && !m_styleStack.isEmpty() )
        {
            return handleDiv(doc, newLine );
        }
        
        return null;
    }
    
    private Element handleBar(Document doc, boolean newLine )
    {
        Element el = null;
        
        if( !m_istable && !newLine )
        {
            return null;
        }

        if( newLine )
        {
            if( !m_istable )
            {
                startBlockLevel();
                el = doc.createElement("table");
                el.setAttribute("class","wikitable");
                el.setAttribute("border","1");
                el = pushElement( el );
                m_istable = true;
                m_rowNum = 0;
            }
            Element tr = doc.createElement("tr");
            el = pushElement( tr );
        }
        
        int ch = nextToken();
        if (ch < 0) return null;

        if( ch == '|' )
        {
            if( !newLine ) 
            {
                el = popElement("th");
            }
            el = pushElement( doc.createElement("th") );
        }
        else
        {
            if( !newLine ) 
            {
                el = popElement("td");
            }
            
            el = pushElement( doc.createElement("td") );

            pushBack( ch );
        }

        return el;
    }

    /**
     *  Generic escape of next character or entity.
     */
    private Element handleTilde()
    {
        int ch = nextToken();
        if (ch < 0) return null;

        if( ch == ' ' )
            return m_currentElement;
        
        if( ch == '|' || ch == '~' || ch == '\\' || ch == '*' || ch == '#' || 
            ch == '-' || ch == '!' || ch == '\'' || ch == '_' || ch == '[' ||
            ch == '{' || ch == ']' || ch == '}' || ch == '%' )
        {
            m_plainTextBuf.append( (char)ch );
            m_plainTextBuf.append(readWhile( ""+(char)ch ));
            return m_currentElement;
        }
 
        // No escape.
        pushBack( ch );

        return null;
    }
       
    private void fillBuffer( Document doc, Element startElement )
    {
        m_currentElement = startElement;
        
        boolean quitReading = false;
        boolean newLine     = true; // FIXME: not true if reading starts in middle of buffer

        disableOutputEscaping(doc);
        
        while(!quitReading)
        {
            int ch = nextToken();
            if (ch < 0) break;
            
            Element el = null;
            
            //
            //  Check if we're actually ending the preformatted mode.
            //  We still must do an entity transformation here.
            //
            if( m_isEscaping )
            {
                if( ch == '}' )
                {
                    if( handleClosebrace() == null ) m_plainTextBuf.append( (char) ch );
                }
                else if( ch == -1 )
                {
                    quitReading = true;
                }
                else if( ch == '\r' )
                {
                    // DOS line feeds we ignore.
                }
                else if( ch == '<' )
                {
                    m_plainTextBuf.append( "&lt;" );
                }
                else if( ch == '>' )
                {
                    m_plainTextBuf.append( "&gt;" );
                }
                else if( ch == '&' )
                {
                    m_plainTextBuf.append( "&amp;" );
                }
                else if( ch == '~' )
                {
                    String braces = readWhile("}");
                    if( braces.length() >= 3 )
                    {
                        m_plainTextBuf.append("}}}");
                     
                        braces = braces.substring(3);
                    }
                    else
                    {
                        m_plainTextBuf.append( (char) ch );                        
                    }
                  
                    for( int i = braces.length()-1; i >= 0; i-- )
                    {
                        pushBack(braces.charAt(i));
                    }
                }
                else 
                {
                    m_plainTextBuf.append( (char) ch );
                }

                continue;
            }

            //
            //  An empty line stops a list
            //
            if( newLine && ch != '*' && ch != '#' && ch != ' ' && m_genlistlevel > 0 )
            {
                m_plainTextBuf.append(unwindGeneralList());
            }

            if( newLine && ch != '|' && m_istable )
            {
                popElement("table");
                m_istable = false;
            }

            //
            //  Now, check the incoming token.
            //
            switch( ch )
            {
              case '\r':
                // DOS linefeeds we forget
                continue;

              case '\n':
                //
                //  Close things like headings, etc.
                //

                // FIXME: This is not really very fast
                popElement("dl"); // Close definition lists.
                popElement("h2");
                popElement("h3");
                popElement("h4");
                if( m_istable ) 
                { 
                    popElement("tr");
                }
                
                m_isdefinition = false;

                if( newLine )
                {
                    // Paragraph change.
                    startBlockLevel();

                    //
                    //  Figure out which elements cannot be enclosed inside
                    //  a <p></p> pair according to XHTML rules.
                    //
                    String nextLine = peekAheadLine();
                    if( nextLine.length() == 0 || 
                        (nextLine.length() > 0 &&
                         !nextLine.startsWith("{{{") &&
                         !nextLine.startsWith("----") &&
                         !nextLine.startsWith("%%") &&
                         "*#!;".indexOf( nextLine.charAt(0) ) == -1) )
                    {
                        pushElement( doc.createElement("p") );
                        m_isOpenParagraph = true;
                        
                        if( m_restartitalic )
                        {
                            pushElement( doc.createElement("i") );
                            m_isitalic = true;
                            m_restartitalic = false;
                        }
                        if( m_restartbold )
                        {
                            pushElement( doc.createElement("b") );
                            m_isbold = true;
                            m_restartbold = false;
                        }
                    }
                }
                else
                {
                    m_plainTextBuf.append("\n");
                    newLine = true;
                }
                continue;
                

              case '\\':
                el = handleBackslash(doc);
                break;

              case '_':
                el = handleUnderscore(doc);
                break;
                
              case '\'':
                el = handleApostrophe(doc);
                break;

              case '{':
                el = handleOpenbrace(doc, newLine );
                break;

              case '}':
                el = handleClosebrace();
                break;

              case '-':
                if( newLine )
                    el = handleDash(doc);

                break;

              case '!':
                if( newLine )
                {
                    el = handleHeading(doc);
                }
                break;

              case ';':
                if( newLine )
                {
                    el = handleDefinitionList(doc);
                }
                break;

              case ':':
                if( m_isdefinition )
                {
                    popElement("dt");
                    el = pushElement( doc.createElement("dd") );
                    m_isdefinition = false;
                }
                break;

              case '[':
                el = handleOpenbracket(doc);
                break;

              case '*':
                if( newLine )
                {
                    pushBack('*');
                    el = handleGeneralList(doc);
                }
                break;

              case '#':
                if( newLine )
                {
                    pushBack('#');
                    el = handleGeneralList(doc);
                }
                break;

              case '|':
                el = handleBar(doc, newLine );
                break;

              case '~':
                el = handleTilde();
                break;

              case '%':
                el = handleDiv(doc, newLine );
                break;

              case '/':
                el = handleSlash(doc, newLine );
                break;
                
              case -1:
                quitReading = true;
                continue;
            }

            //
            //   The idea is as follows:  If the handler method returns
            //   an element (el != null), it is assumed that it has been
            //   added in the stack.  Otherwise the character is added
            //   as is to the plaintext buffer.
            //
            //   For the transition phase, if s != null, it also gets
            //   added in the plaintext buffer.
            //
            if( el != null )
            {
                newLine = false;
            }
            else
            {
                m_plainTextBuf.append( (char) ch );
                newLine = false;
            }
        }
        
        popElement("domroot");
    }

    /*
     * The main Entry point
     * 
     */
    public Document parse()
    {
    	Document d = XMLParser.parse("<body/>");
    	Element rootElement = d.getDocumentElement();
    
        fillBuffer( d, rootElement );
        
        
        return d;
    }
    

    /**
     *  Compares two Strings, and if one starts with the other, then
     *  returns null.  Otherwise just like the normal Comparator
     *  for strings.
     *  
     *  @author jalkanen
     *
     *  @since
     */
    private static class StartingComparator implements Comparator
    {
        public int compare( Object arg0, Object arg1 )
        {
            String s1 = (String)arg0;
            String s2 = (String)arg1;
            
            if( s1.length() > s2.length() )
            {
                if( s1.startsWith(s2) && s2.length() > 1 ) return 0;
            }
            else
            {
                if( s2.startsWith(s1) && s1.length() > 1 ) return 0;
            }
                
            return s1.compareTo( s2 );
        }
        
    }
    

}

