package org.cs3.dokuwiki.jsp.client.output;

import org.cs3.dokuwiki.jsp.client.jspparser.JSPWikiMarkupParser;

import com.google.gwt.xml.client.Document;
import com.google.gwt.xml.client.Element;
import com.google.gwt.xml.client.Node;
import com.google.gwt.xml.client.NodeList;
import com.google.gwt.xml.client.Text;

/** 
 * Converts the Input Tree to DokuWiki-Syntax
 * 
 * @author bihler
 *
 */
public class DokuWikiOutput {

	final static int UL = 0;
	final static int OL = 1;
	private int currentListType = UL;
	private int listIndent = 0;
	
	final static String BOLD = "**";
	final static String ITALIC = "//";
	final static String UNDERLINED = "//";
	final static String MONOSPACED = "''";
	final static String BR = "\\\\ ";
	final static String H1 = "\n====== ";
	final static String H2 = "\n===== ";
	final static String H3 = "\n==== ";
	final static String H4 = "\n=== ";
	final static String H5 = "\n=== ";
	final static String HR = "\n\n----\n";
	
	Document doc;
	

	public DokuWikiOutput(Document doc) {
		super();
		this.doc = doc;
	}
	
	public String dump() {
		
		Element start = doc.getDocumentElement();
		
		return dumpElement(start);
	}
	
	private String dumpElement(Node el) {
		String result = "";
		NodeList nl = el.getChildNodes();
		
		if (nl.getLength() == 0) //some empty tags should produce a space
			if (el instanceof Element) {
				String tag = ((Element) el).getTagName().toLowerCase();
                if ("td".equals(tag) || "th".equals(tag))
					return " ";
			}
		
		for (int i = 0; i < nl.getLength(); i++) {
			Node child =  nl.item(i);
			if (child instanceof Text) {
				
				String text = unescape(((Text) child).getData());
				//Special case of line breaks in li elements
				if (text.startsWith("\n  ") && listIndent > 0)
					text = text.substring(2);
				result += text;
			} else if (child instanceof Element) {
				String tag = ((Element) child).getTagName().toLowerCase();
				/*
				 * Process different instructions:
				 */
				if ("b".equals(tag)) 
					result += dumpElement(BOLD, child);
				else if ("i".equals(tag)) 
					result += dumpElement(ITALIC, child);
				else if ("u".equals(tag)) 
					result += dumpElement(UNDERLINED, child);
				else if ("tt".equals(tag)) 
					result += dumpElement(MONOSPACED, child);
				else if ("pre".equals(tag)) 
					result += dumpElement("<code>","</code>", child);
				else if ("br".equals(tag)) 
					result += BR;
				else if ("h1".equals(tag)) 
					result += dumpElement(H1, reverse(H1), child);
				else if ("h2".equals(tag)) 
					result +=  dumpElement(H2, reverse(H2), child);
				else if ("h3".equals(tag)) 
					result +=  dumpElement(H3, reverse(H3), child);
				else if ("h4".equals(tag)) 
					result += dumpElement(H4, reverse(H4), child);
				else if ("h5".equals(tag)) 
					result += dumpElement(H5, reverse(H5), child);
				else if ("hr".equals(tag)) 
					result += HR;
				else if ("ul".equals(tag)) 
					result += dumpList(UL,child);
				else if ("ol".equals(tag)) 
					result += dumpList(OL,child);
				else if ("li".equals(tag))
					result += dumpListElement(child);
				else if ("a".equals(tag)) {
					String target = ((Element) child).getAttribute("href");
					String caption = null;
					if (child.hasChildNodes())
						caption = unescape(child.getFirstChild().toString());
					if (target != null && target.equals(caption)) 
						caption = null;
					result += dumpLink(target,caption);
				}
				else if ("p".equals(tag)) 
					result += dumpElement("\n\n", child);
				else if ("tr".equals(tag))
					result += dumpElement("\n", "| ", child);
				else if ("td".equals(tag))
					result += dumpElement("|","", child);
				else if ("th".equals(tag))
					result += dumpElement("^","", child);
				else { //ignore
					result += dumpElement(child);
				}
				
				
				
				
				
			}
		}
		
		return result;
	}
	
	private String unescape(String s) {
		s = JSPWikiMarkupParser.StringUtils.replace(s,"&apos;", "'");
		s = JSPWikiMarkupParser.StringUtils.replace(s,"&quot;","\"");
		s = JSPWikiMarkupParser.StringUtils.replace(s,"&lt;","<");
		s = JSPWikiMarkupParser.StringUtils.replace(s,"&gt;",">");
		s = JSPWikiMarkupParser.StringUtils.replace(s,"&amp;","&");
		return s;
	}
	
	private String dumpElement(String delimiter, Node el) {
		return delimiter + dumpElement(el) + delimiter;
	}
	private String dumpElement(String delimiter1,String delimiter2, Node el) {
		return delimiter1 + dumpElement(el) + delimiter2;
	}
	private String reverse(String s) {
		String result = "";
		for (int i = s.length()-1; i >=0; i--) {
			result += s.charAt(i);
		}
		return result;
	}

	private String dumpList(int listType, Node el) {
		int beforeType = currentListType;
		currentListType = listType;
		listIndent++;
		String result = dumpElement(el);
		listIndent--;
		currentListType = beforeType;
		return result;
	}
	
	private String dumpListElement(Node el) {
		String prefix = "\n";
		for (int i = 0; i < listIndent; i++) 
			prefix += "  ";
		switch (currentListType) {
			case OL: prefix += "-"; break;
			default: prefix += "*"; break;
		}
		return prefix + dumpElement(el);
	}
	
	private String dumpLink(String target, String caption) {
		String result = target;
		if (result == null)
			result = "";
		try {
			Integer.parseInt(target);
			//Internal link!
			result = "#" + result;
		} catch (Exception ex) {
			// ignore
		}
		if (caption != null)
			result += "|" + caption;
		
		return "[[" + result + "]]";
	}
}
